using UnityEngine;
using System.Collections;
using UnityEditor;
using System;
using System.Xml.Serialization;
using System.IO;
using System.Reflection;

public class DeformHelper : EditorWindow
{
    GameObject selected = null;
    deform_class[] deformerList;
    deform_config[][] configList;

    object[] setVariables;

    bool customToggle = false;

    int deformLSel = 0;
    int configLSel = 0;
    private Vector2 scrollPos;

    private deform_class selectedClass;
    private bool hasDeformer = false;
    //private bool multi = false;

    private bool savingConfig;
    private string savingName;
    private string savingDesc;
    private deform_class savingClass;

    [MenuItem("Window/Deform Helper")]

    static void Init()
    {
        Starter();
    }

    static void ShowWindow()
    {
        Starter();
    }

    static void OnProjectChange()
    {
        Starter();
    }

    static void OnEnable()
    {
        Starter();
    }

    static void Starter()
    {
        DeformHelper window = (DeformHelper)EditorWindow.GetWindow(typeof(DeformHelper));
        window.selected = Selection.activeGameObject;
        window.title = "Deform Helper";
        window.Repaint();
        window.deformerList = window.GetDeformerList();
        window.configList = window.GetConfigList();
        window.OnSelectionChange();
    }

    void OnGUI()
    {
        EditorGUILayout.BeginVertical();
        scrollPos = EditorGUILayout.BeginScrollView(scrollPos, false, false);

        if (deformerList == null || configList == null)
        {
            deformerList = GetDeformerList();
            configList = GetConfigList();
        }
        if (savingConfig)
        {
            savingName = EditorGUILayout.TextField(new GUIContent("Config name"), savingName);
            EditorGUILayout.HelpBox("If a config for this deformer with the same name exists, it will be overwritten", MessageType.Warning, true);

            GUILayout.Label("Description");
            savingDesc = EditorGUILayout.TextArea(savingDesc, GUILayout.Height(50));

            if (GUILayout.Button("Save"))
            {
                if (savingName == null || savingName == "")
                {
                    savingName = "untitled";
                }

                if (savingDesc == null)
                {
                    savingDesc = "";
                }

                SaveConfig(BuildConfig(savingName, savingDesc, savingClass));
                savingConfig = false;

                configList = GetConfigList();
                
                Repaint();
            }
            if (GUILayout.Button("Cancel"))
            {
                savingConfig = false;
            }
        }
        else
        {
            if (selected != null)
            {
                GUILayout.Label(selected.name, EditorStyles.boldLabel);

                if (hasDeformer)
                {
                    GUILayout.Label(selectedClass.name, EditorStyles.boldLabel);

                    //TODO what if multiple deformers are discovered?

                    Component deformer = selected.GetComponent(selectedClass.component);

                    customToggle = EditorGUILayout.Toggle("Custom", customToggle);
                    EditorGUILayout.Space();
                    if (customToggle)
                    {

                        int i = 0;
                        foreach (deform_classVariable v in selectedClass.variable)
                        {
                            //TODO add support for more types
                            /*
                             * image, vector3, sliders with min/maxes
                             * 
                             */
                            if (v.type == deform_classVariableType.boolean)
                            {
                                if (setVariables[i] == null)
                                {

                                    setVariables[i] = GetPropValue(deformer, v.var_name);
                                }
                                setVariables[i] = EditorGUILayout.Toggle(GenerateVariableGUIContent(v), (bool)setVariables[i]);
                            }
                            else if (v.type == deform_classVariableType.@float)
                            {
                                if (setVariables[i] == null)
                                {
                                    setVariables[i] = GetPropValue(deformer, v.var_name);
                                }
                                setVariables[i] = EditorGUILayout.FloatField(GenerateVariableGUIContent(v), (float)setVariables[i]);
                            }
                            else if (v.type == deform_classVariableType.@int)
                            {
                                if (setVariables[i] == null)
                                {
                                    setVariables[i] = GetPropValue(deformer, v.var_name);
                                }
                                setVariables[i] = EditorGUILayout.IntField(GenerateVariableGUIContent(v), (int)setVariables[i]);
                            }
                            //else if(v.type == deform_classVariableType.image)
                            //{
                            //    if(setVariables[i] == null)
                            //    {
                            //        setVariables[i] = GetPropValue(deformer, v.var_name);
                            //    }
                            //    setVariables[i] = EditorGUILayout.ObjectField(GenerateVariableGUIContent(v), (UnityEngine.Object)setVariables[i], typeof(Texture2D));
                            //}
                            i++;
                        }

                        if (GUILayout.Button(new GUIContent("Apply Changes")))
                        {
                            ApplyConfig(BuildTempConfig(), selectedClass, deformer);
                        }
                        if (GUILayout.Button(new GUIContent("Save as config",
                            "This will allow these settings to be applied quickly to additional objects")))
                        {
                            savingConfig = true;
                            savingDesc = "";
                            savingName = "";
                            savingClass = selectedClass;
                        }
                    }
                    else
                    {
                        string[][] listText = GetConfigListText(Array.IndexOf(deformerList, selectedClass));

                        if (listText[0].Length != 0)
                        {
                            configLSel = EditorGUILayout.Popup(configLSel, listText[0]);
                            EditorGUILayout.HelpBox(listText[1][configLSel], MessageType.Info, true);

                            if (GUILayout.Button(new GUIContent("Apply configuration")))
                            {
                                ApplyConfig(configList[Array.IndexOf(deformerList, selectedClass)][configLSel], selectedClass, deformer);
                            }
                        }
                        else
                        {
                            customToggle = true;
                        }
                    }

                }
                else
                {
                    GUILayout.Label("No deformer found", EditorStyles.boldLabel);
                    EditorGUILayout.Space();
                    EditorGUILayout.Space();
                    deformLSel = EditorGUILayout.Popup(deformLSel, GetDeformerListNames());

                    EditorGUILayout.HelpBox(deformerList[deformLSel].description, MessageType.Info, true);

                    if (GUILayout.Button(new GUIContent("Add deformer")))
                    {
                        selected.AddComponent(LookupComponentName(deformLSel));
                        OnSelectionChange();
                    }
                }

            }
            else
            {
                GUILayout.Label("Bad selection", EditorStyles.boldLabel);
            }
        }
        EditorGUILayout.EndScrollView();
        EditorGUILayout.EndVertical();
    }

    void OnSelectionChange()
    {
        Repaint();
        deformLSel = 0;
        if (Selection.objects.Length > 1)
        {
            selected = null;
            //TODO add support later
        }
        else
        {
            selected = Selection.activeGameObject;

            hasDeformer = Selection.gameObjects.Length == 1 ? CheckDeformer() : false;
        }
    }

    bool CheckDeformer()
    {
        //TODO multideformer
        foreach (deform_class dc in deformerList)
        {
            if (CheckDeformer(dc.component))
            {
                selectedClass = dc;
                setVariables = new object[selectedClass.variable.Length];
                return true;
            }
        }
        return false;
    }

    bool CheckDeformer(string componentName)
    {
        Component c = selected.GetComponent(componentName);
        return (c != null) ? true : false;
    }

    deform_class[] GetDeformerList()
    {
        deform_class[] list;
        string directory = "Assets/Editor/DeformHelper Settings/Classes";

        string[] xmlFiles = Directory.GetFiles(directory, "*.xml");
        list = new deform_class[xmlFiles.Length];

        int i = 0;
        foreach (string file in xmlFiles)
        {
            deform_class dc = GetDeformClass(file);

            if (dc != null)
            {
                list[i] = dc;
                i++;
            }
        }

        return list;
    }

    string[] GetDeformerListNames()
    {
        string[] list = new string[deformerList.Length];
        int i = 0;
        foreach (deform_class dc in deformerList)
        {
            list[i] = dc.name;
            i++;
        }
        return list;
    }

    deform_config[][] GetConfigList()
    {
        deform_config[][] list = new deform_config[deformerList.Length][];

        string[] componentNames = new string[deformerList.Length];

        int i = 0;
        foreach (deform_class dc in deformerList)
        {
            componentNames[i] = dc.component;
            list[i] = new deform_config[0];
            i++;
        }

        string directory = "Assets/Editor/DeformHelper Settings/Configs";
        string[] xmlFiles = Directory.GetFiles(directory, "*.xml");

        foreach (string file in xmlFiles)
        {
            deform_config dc = GetDeformConfig(file);

            int compNo = Array.IndexOf(componentNames, dc.component);
            Array.Resize<deform_config>(ref list[compNo], list[compNo].Length + 1);
            
            list[compNo][list[compNo].Length - 1] = dc;
        }

        return list;
    }

    string[][] GetConfigListText(int componentNo)
    {
        int x = configList[componentNo].Length;
        string[][] list = new string[2][];
        list[0] = new string[x];
        list[1] = new string[x];

        int i = 0;
        foreach (deform_config dc in configList[componentNo])
        {
            list[0][i] = dc.name;
            list[1][i] = dc.description;
            i++;
        }

        return list;
    }

    private static string LookupComponentName(int i)
    {
        string[] xmlFiles = Directory.GetFiles(
            "Assets/Editor/DeformHelper Settings/Classes", "*.xml");

        deform_class dc = GetDeformClass(xmlFiles[i]);

        if (dc != null)
        {
            return dc.component;
        }
        else
        {
            return "";
        }
    }

    private static deform_class GetDeformClass(string filename)
    {
        XmlSerializer ser = new XmlSerializer(typeof(deform_class));
        FileStream fs = new FileStream(filename, FileMode.Open);

        deform_class dc = ser.Deserialize(fs) as deform_class;
        fs.Close();
        return dc;
    }

    private static deform_config GetDeformConfig(string filename)
    {
        XmlSerializer ser = new XmlSerializer(typeof(deform_config));
        FileStream fs = new FileStream(filename, FileMode.Open);
        deform_config dc = ser.Deserialize(fs) as deform_config;
        fs.Close();
        return dc;
    }

    static void ApplyConfig(deform_config dc, deform_class dClass, Component deformer)
    {
        int i = 0;
        deform_classVariable[] cVars = dClass.variable;
        foreach (deform_configVariable dcv in dc.variable)
        {
            deform_classVariableType useType;

            useType = FindVar(cVars, dcv.var_name).type;

            if(useType == deform_classVariableType.boolean)
            {
                bool useVal = Convert.ToBoolean(dcv.value);
                SetField<bool>(deformer, dcv.var_name, useVal);
            }
            else if(useType == deform_classVariableType.@float)
            {
                float useVal = (float)Convert.ToDouble(dcv.value);
                SetField<float>(deformer, dcv.var_name, useVal);
            }
            else if(useType == deform_classVariableType.@int)
            {
                int useVal = Convert.ToInt32(dcv.value);
                SetField<int>(deformer, dcv.var_name, useVal);
            }
            i++;
        }
    }

    static deform_classVariable FindVar(deform_classVariable[] list, string name)
    {
        foreach(deform_classVariable dcv in list)
        {
            if (dcv.var_name == name)
            {
                return dcv;
            }
        }
        return null;
    }

    deform_config BuildConfig(string name, string description, deform_class savingClass)
    {
        if (setVariables[0] != null && selected != null)
        {
            deform_config dc = new deform_config();
            dc.name = name;
            dc.component = savingClass.component;
            dc.description = description;

            deform_configVariable[] list = new deform_configVariable[0];
            int i = 0;
            foreach (deform_classVariable classVar in savingClass.variable)
            {
                deform_configVariable dcv = new deform_configVariable();
                dcv.var_name = classVar.var_name;
                dcv.value = setVariables[i].ToString();
                Array.Resize<deform_configVariable>(ref list, list.Length + 1);
                list[i] = dcv;
                i++;
            }
            dc.variable = list;

            return dc;
        }
        else
        {
            return null;
        }
    }

    deform_config BuildTempConfig()
    {
        if (setVariables[0] != null && selected != null)
        {
            deform_config dc = new deform_config();
            dc.component = selectedClass.component;

            deform_configVariable[] list = new deform_configVariable[0];
            int i = 0;
            foreach (deform_classVariable classVar in selectedClass.variable)
            {
                deform_configVariable dcv = new deform_configVariable();
                dcv.var_name = classVar.var_name;
                dcv.value = setVariables[i].ToString();
                Array.Resize<deform_configVariable>(ref list, list.Length + 1);
                list[i] = dcv;
                i++;
            }
            dc.variable = list;

            return dc;
        }
        else
        {
            return null;
        }
    }

    static void SaveConfig(deform_config dConfig)
    {
        string directory = "Assets/Editor/DeformHelper Settings/Configs/";
        string filename = dConfig.component + "_" + dConfig.name + ".xml";
        XmlSerializer ser = new XmlSerializer(typeof(deform_config));
        FileStream fs = new FileStream(directory + filename, FileMode.Create);
        ser.Serialize(fs, dConfig);
        fs.Close();
    }

    static GUIContent GenerateVariableGUIContent(deform_classVariable dcv)
    {
        GUIContent ret = new GUIContent();
        if (dcv.alt_name != null && dcv.alt_name != "")
        {
            ret.text = dcv.alt_name;
        }
        else
        {
            ret.text = dcv.var_name;
        }

        if (dcv.help != null && dcv.help != "")
        {
            ret.tooltip = dcv.help;
        }

        return ret;
    }

    static object GetPropValue(object src, string propName)
    {
        Debug.Log(src.ToString());
        Debug.Log(propName);
        propName = char.ToLower(propName[0]) + propName.Substring(1);
        Type t = src.GetType();
        FieldInfo fi = t.GetField(propName);

        object ret;

        if(fi != null)
        {
            ret = fi.GetValue(src);
        }
        else
        {
            PropertyInfo pi = t.GetProperty(propName);
            ret = pi.GetValue(src, null);
        }

        return ret;
    }

    static void SetField<T>(object src, string propName, T newValue)
    {

        propName = char.ToLower(propName[0]) + propName.Substring(1);
        Type t = src.GetType();
        FieldInfo info = t.GetField(propName);
        if (info != null)
        {
            info.SetValue(src, newValue);
        }
        else
        {
            PropertyInfo pi = t.GetProperty(propName);
            pi.SetValue(src, newValue, null);
        }
    }
}
